/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or any later
 * version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

package org.objectweb.jonas.discovery;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.net.UnknownHostException;

import javax.management.MBeanServer;
import javax.management.remote.JMXServiceURL;

import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * @author <a href="mailto:Takoua.Abdellatif@inria.fr">Takoua Abdellatif </a>
 * @version 1.0
 */
public class DiscoveryComm implements Runnable {
    private int              RECEIVE_BUFFER_SIZE = 1024;
    /**
     * Used to multicast a discovery event on run() execution
     */
    private MulticastSocket  multicastSocket;
    /**
     * Uset to send a dicovery event as response to a discovey message
     */
    private DatagramSocket unicastSocket;
    /**
     * My discovery manager
     */
    private DiscoveryManager dm;
    /**
     * My manager's listening port
     */
    private int port;
    /**
     * My manager's multicast IP address
     */
    private InetAddress destAddress;
    /**
     * Set to false if the thread is stoped
     */
    private boolean notStopped = true;
    /**
     * Time to live for multicatst packets
     */
    private int ttl = 1; // why 1 ??

    private String jonasName = null;
    private String domainName = null;
    private String[] urls = null;

    /**
     * logger
     */
    private static Logger logger = Log.getLogger(Log.JONAS_DISCOVERY_PREFIX);
    /**
     * Constructs a DiscoveryComm associated to the DiscoveryManager
     * @param dm DiscoveryManager to which this thread is associated
     */
    public DiscoveryComm(DiscoveryManager dm) {
        this.port = dm.getListeningPort();
        try {
            this.destAddress = InetAddress.getByName(dm.getListeningIp());
            this.ttl = dm.getTimeToLive();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        }
        this.dm = dm;
    }
    
    /**
     * Creates a MulticastSocket and joins the group of multicas host
     * identified by the InetAddress <code>destAddress</code>
     *
 	 */
    private void join() {
        try {
            multicastSocket = new MulticastSocket(port);
            multicastSocket.setTimeToLive(ttl);
            multicastSocket.joinGroup(destAddress);
            logger.log(BasicLevel.DEBUG, "multicast ip address is "
                    + destAddress);

            logger.log(BasicLevel.DEBUG, "multicast port is " + port);

        } catch (IOException e) {
            logger.log(BasicLevel.ERROR, "io problem");
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * sends (multicasts) a DiscoveryEvent to the group.
     */
    public void sendNotif(DiscEvent msg) {
        try {
            //send it on the multicast address
            //after transforming the object to a datagram
            logger.log(BasicLevel.DEBUG, msg);
            byte[] messageBytes = DiscEvent.objectToBytes(msg);
            multicastSocket.send(new DatagramPacket(messageBytes,
                    messageBytes.length, destAddress, port));
        } catch (IOException e1) {
            logger.log(BasicLevel.ERROR, "DiscoveryComm: Error to send notification");
            e1.printStackTrace();
        }

    }
/**
 * Send response to a DiscoveryMessage
 * @param msg Containes a DiscoveryEvent allowing to inform about the responder
 * (name, state, URLs)
 * @param destAddress the destination address picked up from the request
 * @param port the destination port picked up from the request
 */
    private void sendResponse(DiscEvent msg, InetAddress destAddress, int port) {
        logger.log(BasicLevel.DEBUG, "DiscoveryComm : The message to send is "
                + msg);
        byte[] messageBytes = DiscEvent.objectToBytes(msg);
        if (messageBytes != null) {
            try {
                // send the unicast response to the discovery client
                unicastSocket.send(new DatagramPacket(messageBytes,
                        messageBytes.length, destAddress, port));
            } catch (IOException e) {
                logger.log(BasicLevel.ERROR, "DiscoveryComm: Error to send response to discovery message");
                e.printStackTrace();
            }
        }
    }

    /**
     * Create a discovery event to notify about a state change of the
     * event sender
     * @param state
     *         - RUNNING if the sender notifies that it gets running
     *         - STOPPING if the sender notifies that it stops running
     * @return a Discovery event (notification)
     * @throws Exception
     *           is thrown if the jmx service is not reached.
     */
    public DiscEvent createNotifMessage(String state) throws Exception {
        String theHostName;
        MBeanServer mbeanServer;
        try {
            theHostName = InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
            return null;
        }

        if (!state.equals(DiscEvent.RUNNING)) {
            urls = null;
        }
        // In the case of a notification, the field port is not important since the
        // notifier is not waiting for an acknowledfement.
        DiscEvent resp = new DiscEvent(theHostName, port, jonasName, domainName, urls);
        resp.setState(state);
        return resp;
    }

    /**
     *
     * @see java.lang.Runnable#run()
     */
    public void run()  {
        // Create a DatagramPacket to receive messages
        DatagramPacket datagram = new DatagramPacket(new byte[RECEIVE_BUFFER_SIZE],
                RECEIVE_BUFFER_SIZE);
        // Object reveived in a message
        Object objReceived = null;
        ObjectInputStream in = null;
        
        // Join the group in order to receive multicast messages
        join();
        // Create notification message containing a discovery event with state RUNNING
        DiscEvent msg = null;
        try {
            msg = createNotifMessage(DiscEvent.RUNNING);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR,
                "DiscoveryComm:  Unable to create a notification message");
            e.printStackTrace();
        }
        if (msg != null) {
            // Multicast the message
            sendNotif(msg);
        }
        // Wait for notification messages
        // Create the socket to be used for responding
        try {
            unicastSocket = new DatagramSocket();
        } catch (SocketException e3) {
            e3.printStackTrace();
            return;
        }
        try {
            while (notStopped) {
                multicastSocket.receive(datagram);
                in = new ObjectInputStream(new ByteArrayInputStream(datagram.getData()));
                objReceived = in.readObject();

                if (objReceived != null) {
                    // The DiscEvents are ignored
                    if (objReceived instanceof DiscMessage) {
                        if (objReceived instanceof DiscEvent) {
                            logger.log(BasicLevel.DEBUG,
                                    "This discovery event is ignored " + objReceived);
                        } else {
                            DiscMessage request = (DiscMessage) objReceived;
                            logger.log(BasicLevel.DEBUG,
                                    "A dicovery message is received "
                                    + objReceived);

                            /* do not create  a new discovery event object with state RUNNING but 
                             * reuse the one created above
                             */
                            /*
                            DiscEvent msg = null;
                            try {
                                msg = createNotifMessage(DiscEvent.RUNNING);
                            } catch (Exception e) {
                                logger.log(BasicLevel.ERROR,
                                    "DiscoveryComm:  Unable to create a notification message");
                                e.printStackTrace();
                            }
                            */
                            if (msg != null) {
                            	InetAddress destAddress = InetAddress.getByName(request.getSourceAddress());
                            	int destPort = request.getSourcePort();
                                sendResponse(msg, destAddress, destPort);
                            }
                        }
                    }
                }
            }
        } catch (SocketException e) {
            logger.log(BasicLevel.ERROR, "Socket closed: " + e);
            notStopped = false;
        } catch (IOException e1) {
            e1.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * sends a notification message to notify that the server is stopping.
     *
     */
    public void stop() {
        // send a notification message of type STOPPING
        DiscEvent msg = null;
        try {
            msg = createNotifMessage(DiscEvent.STOPPING);
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (msg != null) {
            sendNotif(msg);
        }
        Thread.interrupted();
    }

    /**
     * @param jonasName The jonasName to set.
     */
    protected void setJonasName(String jonasName) {
        this.jonasName = jonasName;
    }
    /**
     * @param domainNamle The domainNamle to set.
     */
    protected void setDomainName(String domainName) {
        this.domainName = domainName;
    }
    /**
     * @param urls The urls to set.
     */
    protected void setUrls(String[] urls) {
        this.urls = urls;
    }
}